@@ -484,8 +484,9 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
     int wide;
     int kern_part, rd_part;
     char kern_name[128], rd_name[128];
-    char kern_fullname[128];
+    char kern_fullname[128+10];
     int ok = 1;
+    const char *str;
 
     /* BSS clear */
     bzero(&_edata, &_end - &_edata);
@@ -527,8 +528,14 @@ iplmain(int is_interactive, char *initialstackptr, int started_wide)
 	print_ptab_pretty(partition, sizeof partition / sizeof partition[0]);
     }
 
-    printf("\n%s contains:\n",
-	partitioned ? "PALO(F0) partition" : "Boot image");
+    if (partitioned && f.ipl_addr < partition[0].start * 512)
+	str = "PALO within boot label";
+    else if (partitioned)
+	str = "PALO(F0) partition";
+    else
+	str = "Boot image";
+
+    printf("\n%s contains:\n", str);
 
     if(partitioned && f.version >= 4 && (f.flags & PFLAG_EXT2)) {
 	printf("PALO is formatted EXT2/3\n");
